@@ -3,18 +3,19 @@ from __future__ import annotations
 import json
 from collections import deque
 from decimal import Decimal
+from functools import cached_property
 from os import PathLike
-from typing import Sequence, Mapping, TypeVar, Type, Optional, Iterator, Union, Any, List, Dict
+from typing import Sequence, Mapping, Type, Optional, Iterator, Union, Any, List, Dict
 
 from jschon.jsonpointer import JSONPointer
 from jschon.utils import json_loadf, json_loads
 
 __all__ = [
     'JSON',
-    'AnyJSONCompatible',
+    'JSONCompatible',
 ]
 
-AnyJSONCompatible = TypeVar('AnyJSONCompatible', 'None', bool, int, float, Decimal, str, Sequence, Mapping)
+JSONCompatible = Union[None, bool, int, float, Decimal, str, Sequence[Any], Mapping[str, Any]]
 """Type hint for a JSON-compatible Python object."""
 
 
@@ -41,7 +42,7 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
 
     def __init__(
             self,
-            value: AnyJSONCompatible,
+            value: JSONCompatible,
             *,
             parent: JSON = None,
             key: str = None,
@@ -110,27 +111,25 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
             self.data = value
 
         elif isinstance(value, Sequence):
+            itemclass = itemclass or JSON
             self.type = "array"
-            self.data = []
-            itemclass = itemclass or JSON
-            for i, v in enumerate(value):
-                if isinstance(v, JSON):
-                    v = v.data
-                self.data += [itemclass(v, parent=self, key=str(i), **itemkwargs)]
+            self.data = [
+                itemclass(v, parent=self, key=str(i), **itemkwargs)
+                for i, v in enumerate(value)
+            ]
 
-        elif isinstance(value, Mapping) and all(isinstance(k, str) for k in value):
-            self.type = "object"
-            self.data = {}
+        elif isinstance(value, Mapping):
             itemclass = itemclass or JSON
-            for k, v in value.items():
-                if isinstance(v, JSON):
-                    v = v.data
-                self.data[k] = itemclass(v, parent=self, key=k, **itemkwargs)
+            self.type = "object"
+            self.data = {
+                k: itemclass(v, parent=self, key=k, **itemkwargs)
+                for k, v in value.items()
+            }
 
         else:
             raise TypeError(f"{value=} is not JSON-compatible")
 
-    @property
+    @cached_property
     def path(self) -> JSONPointer:
         """Return the path to the instance from the document root."""
         keys = deque()
@@ -140,8 +139,8 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
             node = node.parent
         return JSONPointer(keys)
 
-    @property
-    def value(self) -> AnyJSONCompatible:
+    @cached_property
+    def value(self) -> JSONCompatible:
         """Return the instance data as a JSON-compatible Python object."""
         if isinstance(self.data, list):
             return [item.value for item in self.data]
@@ -150,10 +149,11 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
         return self.data
 
     def __repr__(self) -> str:
-        # ugly but useful!
+        """Return `repr(self)`."""
         return f'{self.__class__.__name__}({json.loads(str(self))!r})'
 
     def __str__(self) -> str:
+        """Return `str(self)`."""
         def default(o):
             if isinstance(o, JSON):
                 return o.data
@@ -165,23 +165,29 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
                           ensure_ascii=False, allow_nan=False)
 
     def __bool__(self) -> bool:
+        """Return `bool(self)`."""
         return bool(self.data)
 
     def __len__(self) -> int:
+        """Return `len(self)` for an instance of type ``"string"``, ``"array"``
+        or ``"object"``."""
         return len(self.data)
 
     def __iter__(self) -> Iterator:
+        """Return `iter(self)` for an instance of type ``"array"`` or ``"object"``."""
         if self.type in ("array", "object"):
             return iter(self.data)
         raise TypeError(f"{self!r} is not iterable")
 
     def __getitem__(self, index: Union[int, slice, str]) -> JSON:
+        """Return `self[index]` for an instance of type ``"array"`` or ``"object"``."""
         if isinstance(index, (int, slice)) and self.type == "array" or \
                 isinstance(index, str) and self.type == "object":
             return self.data[index]
-        raise TypeError(f"{self!r} is not subscriptable")
+        raise TypeError(f"{self!r} is not subscriptable by {index!r}")
 
-    def __eq__(self, other: Union[JSON, AnyJSONCompatible]) -> bool:
+    def __eq__(self, other: Union[JSON, JSONCompatible]) -> bool:
+        """Return `self == other`."""
         if not isinstance(other, JSON):
             other = JSON(other)
         if self.type == other.type:
@@ -194,30 +200,34 @@ class JSON(Sequence['JSON'], Mapping[str, 'JSON']):
             return self.data == other.data
         return NotImplemented
 
-    def __ge__(self, other: Union[JSON, AnyJSONCompatible]) -> bool:
-        if not isinstance(other, JSON):
-            other = JSON(other)
-        if self.type == other.type:
+    def __ge__(self, other: Union[JSON, int, float, Decimal, str]) -> bool:
+        """Return `self >= other`, for instances of type ``"number"`` or ``"string"``."""
+        if isinstance(other, JSON):
             return self.data >= other.data
-        return NotImplemented
+        if isinstance(other, float):
+            return self.data >= Decimal(f'{other}')
+        return self.data >= other
 
-    def __gt__(self, other: Union[JSON, AnyJSONCompatible]) -> bool:
-        if not isinstance(other, JSON):
-            other = JSON(other)
-        if self.type == other.type:
+    def __gt__(self, other: Union[JSON, int, float, Decimal, str]) -> bool:
+        """Return `self > other`, for instances of type ``"number"`` or ``"string"``."""
+        if isinstance(other, JSON):
             return self.data > other.data
-        return NotImplemented
+        if isinstance(other, float):
+            return self.data > Decimal(f'{other}')
+        return self.data > other
 
-    def __le__(self, other: Union[JSON, AnyJSONCompatible]) -> bool:
-        if not isinstance(other, JSON):
-            other = JSON(other)
-        if self.type == other.type:
+    def __le__(self, other: Union[JSON, int, float, Decimal, str]) -> bool:
+        """Return `self <= other`, for instances of type ``"number"`` or ``"string"``."""
+        if isinstance(other, JSON):
             return self.data <= other.data
-        return NotImplemented
+        if isinstance(other, float):
+            return self.data <= Decimal(f'{other}')
+        return self.data <= other
 
-    def __lt__(self, other: Union[JSON, AnyJSONCompatible]) -> bool:
-        if not isinstance(other, JSON):
-            other = JSON(other)
-        if self.type == other.type:
+    def __lt__(self, other: Union[JSON, int, float, Decimal, str]) -> bool:
+        """Return `self < other`, for instances of type ``"number"`` or ``"string"``."""
+        if isinstance(other, JSON):
             return self.data < other.data
-        return NotImplemented
+        if isinstance(other, float):
+            return self.data < Decimal(f'{other}')
+        return self.data < other
